import logging
from typing import Any, Literal, Optional

from ..colours.get_default_colours import get_default_colours
from ..colours.get_logger_colour import get_logger_colour
from ..logs.log_args import log_args

logger = logging.getLogger("penguin")


class DataProcessing:
    """
    A DataProcessing object handles the pre-processing and post-processing of the function.
    This entails the getting colours, logs, and formatting the logging output.
    """

    def __init__(
        self,
        func: Any,
        args: Any,
        kwargs: Any,
        verbose: Optional[bool] = False,
        show_args: Optional[bool] = False,
        foreground: Optional[
            Literal["red" "yellow", "green", "blue", "magenta", "cyan", "grey"]
        ] = "grey",
        background: Optional[
            Literal[
                "red" "yellow",
                "green",
                "blue",
                "magenta",
                "cyan",
                "grey",
                "black",
                "white",
            ]
        ] = "black",
    ):
        self.func = func
        self.args = args
        self.kwargs = kwargs
        self.verbose = verbose if verbose is not None else False
        self.show_args = show_args if show_args is not None else False
        self.foreground = foreground if foreground is not None else "grey"
        self.background = background if background is not None else "black"
        return

    def pre(self):
        """
        Performs pre-processing on the function:
        It runs common functions used, and gets the foreground colour, background colour, and the function name to run in the @penguin and @penguin_async decorators.
        """
        self.func_name = self.func.__name__
        self.foreground_colour = get_logger_colour(
            self.foreground, "foreground_colours"
        )
        self.background_colour = get_logger_colour(
            self.background, "background_colours"
        )

        if self.show_args or self.verbose:
            log_args(
                self.args,
                self.kwargs,
                self.func_name,
                self.foreground_colour,
                self.background_colour,
            )

        return

    def post(
        self,
        time_msg: str,
        value: Optional[bool] = None,
        verbose: Optional[bool] = False,
        show_return: Optional[bool] = False,
    ):
        """
        Performs post-processing on the function:
        It runs common functions in the @penguin and @penguin_async decorators.
        """
        grey_foreground_colour, black_background_colour = get_default_colours()
        format_start = f"{self.foreground_colour}{self.background_colour}"
        format_end = f"{grey_foreground_colour}{black_background_colour}"
        logger.info(
            f"Finished {format_start}{self.func_name}{format_end} in {format_start}{time_msg}{format_end}"
        )

        if show_return or verbose:
            logger.info(f"Returned value: {format_start}{value!r}{format_end}")

        return



import logging

logger = logging.getLogger("penguin")


def get_time_msg(run_time: float):
    """
    Calculate execution time and prints into readable time string.
    """
    if run_time < 0.0001:
        time_msg = f"{(run_time * 1_000_000):.4f} \u03BCs (microseconds)"
    elif run_time < 1:
        time_msg = f"{(run_time * 1000):.4f} ms (milliseconds)"
    elif run_time < 60:
        time_msg = f"{run_time:.4f} s (seconds)"
    else:
        run_time_mins = run_time / 60
        time_msg = f"{run_time_mins:.4f} min (minutes)"

    return time_msg



from .colour_map import colour_map


def get_default_colours():
    """
    Gets default colours to display in the terminal.

    These being:
    - Foreground: grey
    - Background: black
    """
    grey_foreground_colour = colour_map["foreground_colours"]["grey"]
    black_background_colour = colour_map["background_colours"]["black"]

    return grey_foreground_colour, black_background_colour



import logging
from typing import Any, Tuple

from ..colours.get_default_colours import get_default_colours

logger = logging.getLogger("penguin")


def log_args(
    args: Tuple[Any, ...],
    kwargs: dict,
    func_name: str,
    foreground_colour: str,
    background_colour: str,
):
    """
    Logs the args and kwargs that a function was called with.
    """
    grey_foreground_colour, black_background_colour = get_default_colours()
    args_list = [repr(arg) for arg in args]
    kwargs_list = [f"{k}={v!r}" for k, v, in kwargs.items()]
    params_list = ", ".join([*args_list, *kwargs_list])
    function_signature = f"{func_name}({params_list})"
    format_start = f"{foreground_colour}{background_colour}"
    format_end = f"{grey_foreground_colour}{black_background_colour}"
    logger.info(f"Running {format_start}{function_signature}{format_end}")
    return



from typing import Literal

from .colour_map import colour_map


def get_logger_colour(
    colour: str, colour_type: Literal["background_colours", "foreground_colours"]
):
    """
    Gets the logger foreground and background colours using the colour_map dictionary in colour_map.py
    """
    all_possible_colours = list(colour_map[colour_type].keys())

    if colour in all_possible_colours:
        return colour_map[colour_type][colour]

    colour_dict = colour_map[colour_type]

    if colour_type == "foreground_colours":
        return colour_dict["grey"]
    else:
        return colour_dict["black"]



import logging
import time
from functools import wraps
from typing import Callable, Literal, Optional

from colorama import just_fix_windows_console

from .processing.get_time_msg import get_time_msg
from .processing.processing import DataProcessing

logging.basicConfig(
    level=logging.NOTSET,
    format="%(asctime)s | %(name)s - %(levelname)s - %(message)s",
)
logger = logging.getLogger("penguin")


"""
TODO:
- Update README.md and README_PYPI.md
    - Show async decorator
    - If the function throws an error, the logs won't show up
    - Make more appealing, center the images

"""

"""
Instructions to update new version:

- Delete all files in the dist folder.

- Update the version number in setup.py, README.md, and README_PYPI.md

- Re-create the wheels:
python3 setup.py sdist bdist_wheel

- Re-upload the new files:
twine upload dist/*
"""

just_fix_windows_console()


def penguin(
    verbose: Optional[bool] = False,
    show_args: Optional[bool] = False,
    show_return: Optional[bool] = False,
    foreground: Optional[
        Literal["red" "yellow", "green", "blue", "magenta", "cyan", "grey"]
    ] = "grey",
    background: Optional[
        Literal[
            "red" "yellow", "green", "blue", "magenta", "cyan", "grey", "black", "white"
        ]
    ] = "black",
):
    """
    ## A customizable stopwatch decorator for synchronous functions

    Penguin is a lightweight, customizable decorator that helps you determine how long it takes for your functions to run.

    kwargs:
    `verbose`: When `True`, it shows all logs that are described by the other kwargs. When `False`,
    each kwarg would determine if that specific log is shown
        - default: `False`
    `show_args`: When `True`, it shows the function's signature, with the `*args` and `**kwargs` being passed in.
        - default: `False`
    `show_return`: When `True`, it shows the function's return value(s).
        - default: `False`
    `foreground`: When chosen from this list, `["red" "yellow", "green", "blue", "magenta", "cyan", "grey"]`, it colour the logger output **text** the chosen colour.
        - default: `"grey"`
    `background`: When chosen from this list, `["red" "yellow", "green", "blue", "magenta", "cyan", "grey", "black", "white"]`, it colour the logger output **background** the chosen colour.
        - default: `"black"`
    """

    def penguin_decorator(func: Callable):
        """Log the runtime of the decorated function"""

        @wraps(func)
        def penguin_wrapped(*args, **kwargs):
            data_processing = DataProcessing(
                func,
                args,
                kwargs,
                verbose=verbose,
                show_args=show_args,
                foreground=foreground,
                background=background,
            )
            data_processing.pre()

            try:
                start_time = time.perf_counter()
                value = func(*args, **kwargs)

                end_time = time.perf_counter()
                run_time = end_time - start_time
                time_msg = get_time_msg(run_time)

                data_processing.post(
                    time_msg,
                    value,
                    verbose=verbose,
                    show_return=show_return,
                )
                return value
            except Exception as e:
                raise Exception(e)

        return penguin_wrapped

    return penguin_decorator


def penguin_async(
    verbose: Optional[bool] = False,
    show_args: Optional[bool] = False,
    show_return: Optional[bool] = False,
    foreground: Optional[
        Literal["red" "yellow", "green", "blue", "magenta", "cyan", "grey"]
    ] = "grey",
    background: Optional[
        Literal[
            "red" "yellow", "green", "blue", "magenta", "cyan", "grey", "black", "white"
        ]
    ] = "black",
):
    """
    ## A customizable stopwatch decorator for coroutines (asynchronous functions)

    Penguin is a lightweight, customizable decorator that helps you determine how long it takes for your functions to run.

    kwargs:
    `verbose`: When `True`, it shows all logs that are described by the other kwargs. When `False`,
    each kwarg would determine if that specific log is shown
        - default: `False`
    `show_args`: When `True`, it shows the function's signature, with the `*args` and `**kwargs` being passed in.
        - default: `False`
    `show_return`: When `True`, it shows the function's return value(s).
        - default: `False`
    `foreground`: When chosen from this list, `["red" "yellow", "green", "blue", "magenta", "cyan", "grey"]`, it colour the logger output **text** the chosen colour.
        - default: `"grey"`
    `background`: When chosen from this list, `["red" "yellow", "green", "blue", "magenta", "cyan", "grey", "black", "white"]`, it colour the logger output **background** the chosen colour.
        - default: `"black"`
    """

    def penguin_decorator(func: Callable):
        """Log the runtime of the decorated function"""

        @wraps(func)
        async def penguin_wrapped(*args, **kwargs):
            data_processing = DataProcessing(
                func,
                args,
                kwargs,
                verbose=verbose,
                show_args=show_args,
                foreground=foreground,
                background=background,
            )
            data_processing.pre()

            try:
                start_time = time.perf_counter()
                value = await func(*args, **kwargs)

                end_time = time.perf_counter()
                run_time = end_time - start_time
                time_msg = get_time_msg(run_time)

                data_processing.post(
                    time_msg,
                    value,
                    verbose=verbose,
                    show_return=show_return,
                )
                return value
            except Exception as e:
                raise Exception(e)

        return penguin_wrapped

    return penguin_decorator



import logging
from typing import Any

logger = logging.getLogger("penguin")


def log_return(value: Any):
    """
    Logs the return value of the function that was called.
    """
    logger.info(f"Returned value: {value!r}")
    return
